from __future__ import annotations

import email
import mimetypes
import os
from typing import TYPE_CHECKING

import httpx
from httpx._config import DEFAULT_LIMITS
from httpx._content import IteratorByteStream

from unearth.link import Link
from unearth.utils import parse_netloc

if TYPE_CHECKING:
    from typing import Any, ContextManager, Iterable, Mapping

    from httpx._types import CertTypes, TimeoutTypes, VerifyTypes


def is_absolute_url(self) -> bool:
    return self._uri_reference.scheme or self._uri_reference.host


# Patch the is_absolute_url method of httpx.URL to allow file:// URLs
httpx.URL.is_absolute_url = property(is_absolute_url)


class FileByteStream(IteratorByteStream):
    def close(self) -> None:
        self._stream.close()  # type: ignore[attr-defined]


class LocalFSTransport(httpx.BaseTransport):
    def handle_request(self, request: httpx.Request) -> httpx.Response:
        link = Link(str(request.url))
        path = link.file_path
        if request.method != "GET":
            return httpx.Response(status_code=405)

        try:
            stats = os.stat(path)
        except OSError as exc:
            # format the exception raised as a io.BytesIO object,
            # to return a better error message:
            return httpx.Response(status_code=404, text=f"{type(exc).__name__}: {exc}")
        else:
            modified = email.utils.formatdate(stats.st_mtime, usegmt=True)
            content_type = mimetypes.guess_type(path)[0] or "text/plain"
            headers = {
                "Content-Type": content_type,
                "Content-Length": str(stats.st_size),
                "Last-Modified": modified,
            }
            return httpx.Response(
                status_code=200,
                headers=headers,
                stream=FileByteStream(path.open("rb")),
            )


class PyPIClient(httpx.Client):
    """
    A :class:`httpx.Client` subclass that supports file:// URLs and trusted hosts configuration.

    Args:
        trusted_hosts: A list of trusted hosts. If a host is trusted, the client will not verify the SSL certificate.
        \\**kwargs: Additional keyword arguments to pass to the :class:`httpx.Client` constructor.
    """

    def __init__(
        self,
        *,
        trusted_hosts: Iterable[str] = (),
        verify: VerifyTypes = True,
        cert: CertTypes | None = None,
        http1: bool = True,
        http2: bool = False,
        limits: httpx.Limits = DEFAULT_LIMITS,
        trust_env: bool = True,
        timeout: TimeoutTypes = 10.0,
        **kwargs: Any,
    ) -> None:
        self._trusted_host_ports: set[tuple[str, int | None]] = set()
        # Due to lack of ability of retry behavior in httpx, we don't support it for simplicity
        insecure_transport = httpx.HTTPTransport(
            verify=False,
            cert=cert,
            http1=http1,
            http2=http2,
            limits=limits,
            trust_env=trust_env,
        )

        mounts: dict[str, httpx.BaseTransport] = {"file://": LocalFSTransport()}
        for host in trusted_hosts:
            hostname, port = parse_netloc(host)
            self._trusted_host_ports.add((hostname, port))
            mounts[f"all://{host}"] = insecure_transport

        mounts.update(kwargs.pop("mounts", {}))

        super().__init__(
            verify=verify,
            cert=cert,
            http1=http1,
            http2=http2,
            limits=limits,
            trust_env=trust_env,
            timeout=timeout,
            mounts=mounts,
            **kwargs,
        )

    def get_stream(
        self, url: str, *, headers: Mapping[str, str] | None = None
    ) -> ContextManager[httpx.Response]:
        return self.stream("GET", url, headers=headers)

    def iter_secure_origins(self) -> Iterable[tuple[str, str, str]]:
        from unearth.fetchers import DEFAULT_SECURE_ORIGINS

        yield from DEFAULT_SECURE_ORIGINS
        for host, port in self._trusted_host_ports:
            yield ("*", host, "*" if port is None else str(port))



"""Configuration for the pytest test suite."""

from __future__ import annotations

import os
import sys
from ssl import SSLContext
from typing import TYPE_CHECKING

import flask
import pytest
from httpx import WSGITransport
from wsgiadapter import WSGIAdapter as _WSGIAdapter

from tests.fixtures.app import BASE_DIR, create_app
from unearth.fetchers import PyPIClient
from unearth.fetchers.legacy import InsecureMixin, PyPISession

if TYPE_CHECKING:
    from typing import Literal


class WSGIAdapter(_WSGIAdapter):
    def send(self, request, *args, **kwargs):
        resp = super().send(request, *args, **kwargs)
        resp.connection = self
        return resp


class InsecureWSGIAdapter(InsecureMixin, WSGIAdapter):
    pass


@pytest.fixture(scope="session")
def custom_certificate_authority():
    if sys.version_info >= (3, 13):
        pytest.skip("trustme is not compatible with Python 3.13")
    import trustme

    return trustme.CA()


@pytest.fixture(scope="session")
def self_signed_server_cert(httpserver_listen_address, custom_certificate_authority):
    server, _ = httpserver_listen_address
    return custom_certificate_authority.issue_cert(server)


@pytest.fixture(scope="session")
def httpserver_ssl_context(self_signed_server_cert):
    server_context = SSLContext()
    self_signed_server_cert.configure_cert(server_context)
    return server_context


@pytest.fixture(scope="session")
def httpserver_listen_address():
    return (
        "localhost",
        # select the port randomly
        0,
    )


@pytest.fixture()
def fixtures_dir():
    return BASE_DIR


@pytest.fixture()
def pypi():
    return create_app()


@pytest.fixture()
def pypi_auth(pypi):
    def check_auth(auth):
        return auth.username == os.getenv(
            "PYPI_USER", "test"
        ) and auth.password == os.getenv("PYPI_PASSWORD", "password")

    def unauthenticated():
        message = {"message": "Unauthenticated"}
        resp = flask.make_response(flask.jsonify(message), 401)
        resp.headers["WWW-Authenticate"] = "Basic realm='Main'"
        return resp

    @pypi.before_request
    def require_basic_auth():
        auth = flask.request.authorization
        if not auth or not check_auth(auth):
            return unauthenticated()

    return pypi


@pytest.fixture(params=["sync", "legacy"])
def fetcher_type(request) -> Literal["sync", "legacy"]:
    return request.param


@pytest.fixture()
def session(fetcher_type):
    if fetcher_type == "sync":
        client = PyPIClient()
    else:
        client = PyPISession()
    try:
        yield client
    finally:
        client.close()


@pytest.fixture()
def pypi_session(pypi, fetcher_type, mocker):
    if fetcher_type == "sync":
        client = PyPIClient(transport=WSGITransport(pypi))
    else:
        mocker.patch.object(
            PyPISession, "insecure_adapter_cls", return_value=InsecureWSGIAdapter(pypi)
        )
        mocker.patch.object(
            PyPISession, "secure_adapter_cls", return_value=WSGIAdapter(pypi)
        )
        client = PyPISession()
    try:
        yield client
    finally:
        client.close()


@pytest.fixture(params=["html", "json"])
def content_type(request, monkeypatch) -> str:
    monkeypatch.setenv("INDEX_RETURN_TYPE", request.param)
    return request.param



import pytest
from packaging.requirements import Requirement

from unearth.evaluator import (
    Evaluator,
    FormatControl,
    Package,
    TargetPython,
    evaluate_package,
    validate_hashes,
)
from unearth.link import Link

BINARY_LINKS = [
    Link("https://test.pypi.org/files/click-8.1.3-py3-none-any.whl"),
    Link("file:///home/user/click-8.1.3-py3-none-any.whl"),
]

SOURCE_LINKS = [
    Link("file:///home/user/code/click"),
    Link("https://test.pypi.org/files/click-8.1.3.tar.gz"),
    Link("https://test.pypi.org/files/Jinja2-3.1.2.zip"),
    Link("git+https://github.com/pallets/click.git@main"),
]


@pytest.mark.parametrize("link", BINARY_LINKS)
def test_only_binary_is_allowed(link):
    format_control = FormatControl(only_binary={"foo"})
    format_control.check_format(link, "foo")

    format_control = FormatControl(no_binary={"foo"})
    with pytest.raises(ValueError):
        format_control.check_format(link, "foo")


@pytest.mark.parametrize("link", SOURCE_LINKS)
def test_no_binary_is_allowed(link):
    format_control = FormatControl(only_binary={"foo"})
    with pytest.raises(ValueError):
        format_control.check_format(link, "foo")

    format_control = FormatControl(no_binary={"foo"})
    format_control.check_format(link, "foo")


@pytest.mark.parametrize("link", BINARY_LINKS + SOURCE_LINKS)
def test_default_format_control_allow_all(link):
    format_control = FormatControl()
    format_control.check_format(link, "foo")


@pytest.mark.parametrize("allow_yanked", (True, False))
def test_evaluate_yanked_link(allow_yanked):
    link = Link(
        "https://test.pypi.org/files/click-8.1.3-py3-none-any.whl", yank_reason="bad"
    )
    evaluator = Evaluator("click", allow_yanked=allow_yanked)
    if allow_yanked:
        assert evaluator.evaluate_link(link) is not None
    else:
        assert evaluator.evaluate_link(link) is None


@pytest.mark.parametrize(
    "python_version,requires_python,expected",
    [
        ((3, 9), None, True),
        ((3, 9), ">=3.9", True),
        ((3, 8), ">=3.9", False),
    ],
)
@pytest.mark.parametrize("ignore_compatibility", (True, False))
def test_evaluate_link_python_version(
    python_version, requires_python, expected, ignore_compatibility
):
    link = Link(
        "https://test.pypi.org/files/click-8.1.3-py3-none-any.whl",
        requires_python=requires_python,
    )
    evaluator = Evaluator(
        "click",
        target_python=TargetPython(python_version),
        ignore_compatibility=ignore_compatibility,
    )
    assert (evaluator.evaluate_link(link) is None) is not (
        expected or ignore_compatibility
    )


@pytest.mark.parametrize(
    "url",
    [
        "https://test.pypi.org/files/click-8.1.3.whl",
        "https://test.pypi.org/files/click-8.1.3_develop-py3-none-any.whl",
    ],
)
def test_evaluate_invalid_wheel_name(url):
    link = Link(url)
    evaluator = Evaluator("click")
    assert evaluator.evaluate_link(link) is None


@pytest.mark.parametrize("loose", (True, False))
def test_evaluate_link_loose_filename(loose: bool, monkeypatch: pytest.MonkeyPatch):
    if loose:
        monkeypatch.setenv("UNEARTH_LOOSE_FILENAME", "1")
    link = Link("https://test.pypi.org/files/foo-2-2.tar.gz")
    evaluator = Evaluator("foo")
    package = evaluator.evaluate_link(link)
    if loose:
        assert package is not None and package.version == "2-2"
    else:
        assert package is None


@pytest.mark.parametrize(
    "link,expected",
    [
        ("https://test.pypi.org/files/click-8.1.3-py3-none-any.whl", True),
        ("https://test.pypi.org/files/Click-8.1.3.tar.gz", True),
        ("https://test.pypi.org/files/Jinja2-3.1.2.zip", False),
    ],
)
def test_evaluate_against_name_match(link, expected):
    evaluator = Evaluator("click")
    assert (evaluator.evaluate_link(Link(link)) is None) is not expected


@pytest.mark.parametrize(
    "link",
    [
        Link("https://test.pypi.org/files/click.zip"),
        Link("https://test.pypi.org/files/click.tar.gz"),
        Link("git+git@github.com:pallets/click.git@main"),
        Link("git+git@github.com:pallets/click.git@main#egg=click"),
    ],
)
def test_evaluate_against_missing_version(link):
    evaluator = Evaluator("click")
    assert evaluator.evaluate_link(link) is None


@pytest.mark.parametrize(
    "url,match",
    [
        (
            "https://test.pypi.org/files/click-8.1.3-py3-none-any.whl"
            "#sha256=1234567890abcdef",
            True,
        ),
        (
            "https://test.pypi.org/files/click-8.1.3-py3-none-any.whl"
            "#sha256=fedcba0987654321",
            True,
        ),
        (
            "https://test.pypi.org/files/click-8.1.3-py3-none-any.whl"
            "#sha256=1112222",
            False,
        ),
    ],
)
def test_evaluate_against_allowed_hashes(url, match, session):
    package = Package("click", "8.1.3", link=Link(url))
    result = validate_hashes(
        package,
        {"sha256": ["1234567890abcdef", "fedcba0987654321"]},
        session=session,
    )
    assert result is match


@pytest.mark.parametrize(
    "url",
    [
        "https://test.pypi.org/files/click-8.1.3-py3-none-any.whl",
        "https://test.pypi.org/files/click-8.1.3-py3-none-any.whl#sha256=123456",
        "https://test.pypi.org/files/click-8.1.3-py3-none-any.whl#md5=1111222",
    ],
)
def test_evaluate_allow_all_hashes(url, session):
    package = Package("click", "8.1.3", link=Link(url))
    assert validate_hashes(package, {}, session)


@pytest.mark.parametrize(
    "url",
    [
        "https://test.pypi.org/files/click-8.1.3-py3-none-any.whl",
        "https://test.pypi.org/files/click-8.1.3-py3-none-any.whl#md5=1111222",
    ],
)
def test_retrieve_hash_from_internet(pypi_session, url):
    link = Link(url)
    package = Package("click", "8.1.3", link=link)
    assert validate_hashes(
        package,
        hashes={
            "sha256": [
                "bb4d8133cb15a609f44e8213d9b391b0809795062913b383c62be0ee95b1db48"
            ]
        },
        session=pypi_session,
    )
    hash_name, hash = next(iter(link.hashes.items()))
    assert hash_name == "sha256"
    assert hash == "bb4d8133cb15a609f44e8213d9b391b0809795062913b383c62be0ee95b1db48"


@pytest.mark.filterwarnings("ignore::FutureWarning")
@pytest.mark.parametrize(
    "link,expected",
    [
        (Link("https://test.pypi.org/files/click-8.1.3-py3-none-any.whl"), True),
        (Link("https://test.pypi.org/files/click-8.1.3-cp39-cp39-win_amd64.whl"), True),
        (
            Link("https://test.pypi.org/files/click-8.1.3-cp310-cp310-win_amd64.whl"),
            False,
        ),
        (
            Link(
                "https://test.pypi.org/files/click-8.1.3-cp39-cp39-"
                "macosx_11_0_arm64.whl"
            ),
            False,
        ),
        (
            Link(
                "https://test.pypi.org/files/click-8.1.3-py3-none-any.whl",
                requires_python=">3.6.*",
            ),
            True,
        ),
    ],
)
@pytest.mark.parametrize("ignore_compatibility", (True, False))
def test_evaluate_compatibility_tags(link, expected, ignore_compatibility):
    evaluator = Evaluator(
        "click",
        target_python=TargetPython((3, 9), ["cp39"], "cp", ["win_amd64"]),
        ignore_compatibility=ignore_compatibility,
    )
    assert (evaluator.evaluate_link(link) is None) is not (
        expected or ignore_compatibility
    )


@pytest.mark.parametrize(
    "version,requires,allow_prereleases,expected",
    [
        ("8.1.3", ">=8.0", None, True),
        ("7.1", ">=8.0", None, False),
        ("8.0.0a0", ">=8.0.0dev0", None, True),
        ("8.0.0dev0", ">=7", None, False),
        ("8.0.0dev0", ">=7", True, True),
        ("8.0.0a0", "", None, False),
        ("8.0.0a0", ">=8.0.0dev0", False, False),
    ],
)
def test_evaluate_packages_matching_version(
    version, requires, allow_prereleases, expected
):
    requirement = Requirement(f"click{requires}")
    link = Link(f"https://test.pypi.org/packages/source/c/click/click-{version}.tar.gz")
    package = Package("click", version, link)
    assert evaluate_package(package, requirement, allow_prereleases) is expected


def test_evaluate_packages_matching_url():
    requirement = Requirement(
        "click @ https://test.pypi.org/packages/source/c/click/click-8.1.3.tar.gz"
    )
    link = Link("https://test.pypi.org/packages/source/c/click/click-8.1.3.tar.gz")
    package = Package("click", None, link)
    assert evaluate_package(package, requirement, None)



"""The main class for the unearth package."""

from __future__ import annotations

import atexit
import functools
import itertools
import os
import pathlib
import posixpath
import warnings
from datetime import datetime
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING, Any, Iterable, NamedTuple, Sequence

import packaging.requirements
from packaging.utils import BuildTag, canonicalize_name, parse_wheel_filename
from packaging.version import parse as parse_version

from unearth.auth import MultiDomainBasicAuth
from unearth.collector import collect_links_from_location
from unearth.evaluator import (
    Evaluator,
    FormatControl,
    Package,
    TargetPython,
    evaluate_package,
    is_equality_specifier,
    validate_hashes,
)
from unearth.fetchers import Fetcher
from unearth.fetchers.sync import PyPIClient
from unearth.link import Link
from unearth.preparer import noop_download_reporter, noop_unpack_reporter, unpack_link
from unearth.utils import LazySequence

if TYPE_CHECKING:
    from typing import TypedDict

    from unearth.preparer import DownloadReporter, UnpackReporter

    class Source(TypedDict):
        url: str
        type: str

else:
    Source = dict


def _check_legacy_session(session: Any) -> None:
    try:
        from requests import Session
    except ModuleNotFoundError:
        return

    if isinstance(session, Session):
        warnings.warn(
            "The legacy requests.Session is used, which is deprecated and will be removed in the next release. "
            "Please use `httpx.Client` instead. ",
            DeprecationWarning,
            stacklevel=2,
        )


class BestMatch(NamedTuple):
    """The best match for a package."""

    #: The best matching package, or None if no match was found.
    best: Package | None
    #: The applicable packages, excluding those with unmatching versions.
    applicable: Sequence[Package]
    #: All candidates found for the requirement.
    candidates: Sequence[Package]


class PackageFinder:
    """The main class for the unearth package.

    Args:
        session: The session to use for the finder.
            If not provided, a temporary session will be created.
        index_urls: The index URLs to search for packages.
        find_links: The links to search for packages.
        trusted_hosts:: The trusted hosts.
        target_python: The links must match
            the target Python
        ignore_compatibility: Whether to ignore the compatibility check
        no_binary: The names of the packages to disallow wheels
        only_binary: The names of the packages to disallow non-wheels
        prefer_binary: The names of the packages to prefer binary
            distributions even if newer sdist pacakges exist.
        respect_source_order: If True, packages from the source coming earlier
            are more preferred, even if they have lower versions.
        verbosity: The verbosity level.
        exclude_newer_than: A datetime that when provided, excludes any version
            newer than it.
    """

    def __init__(
        self,
        session: Fetcher | None = None,
        *,
        index_urls: Iterable[str] = (),
        find_links: Iterable[str] = (),
        trusted_hosts: Iterable[str] = (),
        target_python: TargetPython | None = None,
        ignore_compatibility: bool = False,
        no_binary: Iterable[str] = (),
        only_binary: Iterable[str] = (),
        prefer_binary: Iterable[str] = (),
        respect_source_order: bool = False,
        verbosity: int = 0,
        exclude_newer_than: datetime | None = None,
    ) -> None:
        self.sources: list[Source] = []
        for url in index_urls:
            self.add_index_url(url)
        for url in find_links:
            self.add_find_links(url)
        # Add PyPI as the default index.
        if not self.sources:
            self.add_index_url("https://pypi.org/simple/")
        self.target_python = target_python or TargetPython()
        self.ignore_compatibility = ignore_compatibility
        self.no_binary = {canonicalize_name(name) for name in no_binary}
        self.only_binary = {canonicalize_name(name) for name in only_binary}
        self.prefer_binary = {canonicalize_name(name) for name in prefer_binary}
        self.trusted_hosts = trusted_hosts
        _check_legacy_session(session)
        self._session = session
        self.respect_source_order = respect_source_order
        self.verbosity = verbosity
        self.exclude_newer_than = exclude_newer_than
        self.headers: dict[str, str] = {}

        self._tag_priorities = {
            tag: i for i, tag in enumerate(self.target_python.supported_tags())
        }

    @property
    def session(self) -> Fetcher:
        if self._session is None:
            index_urls = [
                source["url"] for source in self.sources if source["type"] == "index"
            ]
            session = PyPIClient(trusted_hosts=self.trusted_hosts)
            session.auth = MultiDomainBasicAuth(index_urls=index_urls)
            atexit.register(session.close)
            self._session = session
        return self._session

    def add_index_url(self, url: str) -> None:
        """Add an index URL to the finder search scope.

        Args:
            url (str): The index URL to add.
        """
        self.sources.append({"url": url, "type": "index"})

    def add_find_links(self, url: str) -> None:
        """Add a find links URL to the finder search scope.

        Args:
            url (str): The find links URL to add.
        """
        self.sources.append({"url": url, "type": "find_links"})

    def build_evaluator(
        self, package_name: str, allow_yanked: bool = False
    ) -> Evaluator:
        """Build an evaluator for the given package name.

        Args:
            package_name : The desired package name
            allow_yanked: Whether to allow yanked candidates.

        Returns:
            Evaluator: The evaluator for the given package name
        """
        format_control = FormatControl(
            no_binary=self.no_binary, only_binary=self.only_binary
        )
        return Evaluator(
            package_name=package_name,
            target_python=self.target_python,
            ignore_compatibility=self.ignore_compatibility,
            allow_yanked=allow_yanked,
            format_control=format_control,
            exclude_newer_than=self.exclude_newer_than,
        )

    def _build_index_page_link(self, index_url: str, package_name: str) -> Link:
        url = posixpath.join(index_url, canonicalize_name(package_name)) + "/"
        return self._build_find_link(url)

    def _build_find_link(self, find_link: str) -> Link:
        if os.path.exists(find_link):
            return Link.from_path(os.path.abspath(find_link))
        elif "://" in find_link:
            return Link(find_link)
        raise ValueError(f"Invalid find link or non-existing path: {find_link}")

    def _evaluate_links(
        self, links: Iterable[Link], evaluator: Evaluator
    ) -> Iterable[Package]:
        return filter(None, map(evaluator.evaluate_link, links))

    def _evaluate_packages(
        self,
        packages: Iterable[Package],
        requirement: packaging.requirements.Requirement,
        allow_prereleases: bool | None = None,
    ) -> Iterable[Package]:
        evaluator = functools.partial(
            evaluate_package,
            requirement=requirement,
            allow_prereleases=allow_prereleases,
        )
        return filter(evaluator, packages)

    def _evaluate_hashes(
        self, packages: Iterable[Package], hashes: dict[str, list[str]]
    ) -> Iterable[Package]:
        evaluator = functools.partial(
            validate_hashes, hashes=hashes, session=self.session
        )
        return filter(evaluator, packages)

    def _sort_key(self, package: Package) -> tuple:
        """The key for sort, package with the largest value is the most preferred."""
        link = package.link
        pri = len(self._tag_priorities) + 1  # default priority for sdist.
        build_tag: BuildTag = ()
        prefer_binary = False
        if link.is_wheel:
            *_, build_tag, file_tags = parse_wheel_filename(link.filename)
            pri = min(
                (self._tag_priorities.get(tag, pri - 1) for tag in file_tags),
                default=pri - 1,
            )
            if (
                canonicalize_name(package.name) in self.prefer_binary
                or ":all:" in self.prefer_binary
            ):
                prefer_binary = True

        return (
            -int(link.is_yanked),
            int(prefer_binary),
            parse_version(package.version) if package.version is not None else 0,
            -pri,
            build_tag,
        )

    def _find_packages(
        self, package_name: str, allow_yanked: bool = False
    ) -> Iterable[Package]:
        """Find all packages with the given name.

        Args:
            package_name: The desired package name
            allow_yanked: Whether to allow yanked candidates.

        Returns:
            The packages with the given name, sorted by best match.
        """
        evaluator = self.build_evaluator(package_name, allow_yanked)

        def find_one_source(source: Source) -> Iterable[Package]:
            if source["type"] == "index":
                link = self._build_index_page_link(source["url"], package_name)
                result = self._evaluate_links(
                    collect_links_from_location(
                        self.session, link, headers=self.headers
                    ),
                    evaluator,
                )
            else:
                link = self._build_find_link(source["url"])
                result = self._evaluate_links(
                    collect_links_from_location(
                        self.session, link, expand=True, headers=self.headers
                    ),
                    evaluator,
                )
            if self.respect_source_order:
                # Sort the result within the individual source.
                return sorted(result, key=self._sort_key, reverse=True)
            return result

        all_packages = itertools.chain.from_iterable(map(find_one_source, self.sources))
        if self.respect_source_order:
            return all_packages
        # Otherwise, sort the result across all sources.
        return sorted(all_packages, key=self._sort_key, reverse=True)

    def find_all_packages(
        self,
        package_name: str,
        allow_yanked: bool = False,
        hashes: dict[str, list[str]] | None = None,
    ) -> Sequence[Package]:
        """Find all packages with the given package name, best match first.

        Args:
            package_name (str): The desired package name
            allow_yanked (bool): Whether to allow yanked candidates.
            hashes (dict[str, list[str]]|None): The hashes to filter on.

        Returns:
            Sequence[Package]: The packages list sorted by best match
        """
        return LazySequence(
            self._evaluate_hashes(
                self._find_packages(package_name, allow_yanked), hashes=hashes or {}
            )
        )

    def _find_packages_from_requirement(
        self,
        requirement: packaging.requirements.Requirement,
        allow_yanked: bool | None = None,
    ) -> Iterable[Package]:
        if allow_yanked is None:
            allow_yanked = is_equality_specifier(requirement.specifier)
        if requirement.url:
            yield Package(requirement.name, None, link=Link(requirement.url))
        else:
            yield from self._find_packages(requirement.name, allow_yanked)

    def find_matches(
        self,
        requirement: packaging.requirements.Requirement | str,
        allow_yanked: bool | None = None,
        allow_prereleases: bool | None = None,
        hashes: dict[str, list[str]] | None = None,
    ) -> Sequence[Package]:
        """Find all packages matching the given requirement, best match first.

        Args:
            requirement: A packaging.requirements.Requirement
                instance or a string to construct it.
            allow_yanked (bool|None): Whether to allow yanked candidates,
                or None to infer from the specifier.
            allow_prereleases (bool|None): Whether to allow prereleases,
                or None to infer from the specifier.
            hashes (dict[str, list[str]]|None): The hashes to filter on.

        Returns:
            Sequence[Package]: The packages sorted by best match
        """
        if isinstance(requirement, str):
            requirement = packaging.requirements.Requirement(requirement)
        return LazySequence(
            self._evaluate_hashes(
                self._evaluate_packages(
                    self._find_packages_from_requirement(requirement, allow_yanked),
                    requirement,
                    allow_prereleases,
                ),
                hashes=hashes or {},
            )
        )

    def find_best_match(
        self,
        requirement: packaging.requirements.Requirement | str,
        allow_yanked: bool | None = None,
        allow_prereleases: bool | None = None,
        hashes: dict[str, list[str]] | None = None,
    ) -> BestMatch:
        """Find the best match for the given requirement.

        Args:
            requirement: A packaging.requirements.Requirement
                instance or a string to construct it.
            allow_yanked (bool|None): Whether to allow yanked candidates,
                or None to infer from the specifier.
            allow_prereleases (bool|None): Whether to allow prereleases,
                or None to infer from the specifier.
            hashes (dict[str, list[str]]|None): The hashes to filter on.

        Returns:
            BestMatch: The best match
        """
        if isinstance(requirement, str):
            requirement = packaging.requirements.Requirement(requirement)
        packages = self._find_packages_from_requirement(requirement, allow_yanked)
        candidates = LazySequence(packages)
        applicable_candidates = LazySequence(
            self._evaluate_hashes(
                self._evaluate_packages(packages, requirement, allow_prereleases),
                hashes=hashes or {},
            )
        )
        best_match = next(iter(applicable_candidates), None)
        return BestMatch(best_match, applicable_candidates, candidates)

    def download_and_unpack(
        self,
        link: Link,
        location: str | pathlib.Path,
        download_dir: str | pathlib.Path | None = None,
        hashes: dict[str, list[str]] | None = None,
        download_reporter: DownloadReporter = noop_download_reporter,
        unpack_reporter: UnpackReporter = noop_unpack_reporter,
    ) -> pathlib.Path:
        """Download and unpack the package at the given link.

        If the link is a remote link, it will be downloaded to the ``download_dir``.
        Then, if the link is a wheel, the path of wheel file will be returned,
        otherwise it will be unpacked to the destination. Specially, if the link refers
        to a local directory, the path will be returned directly, otherwise the unpacked
        source path will be returned. And if the link has a subdirectory fragment, the
        subdirectory will be returned.

        Args:
            link (Link): The link to download
            location: The destination directory
            download_dir: The directory to download to, or None to use a
                temporary directory created by unearth.
            hashes (dict[str, list[str]]|None): The optional hash dict for validation.
            download_reporter (DownloadReporter): The download reporter for progress
                reporting. By default, it does nothing.
            unpack_reporter (UnpackReporter): The unpack reporter for progress
                reporting. By default, it does nothing.

        Returns:
            The path to the installable file or directory.
        """
        import contextlib

        if hashes is None:
            hashes = link.hash_option

        with contextlib.ExitStack() as stack:
            if download_dir is None:
                download_dir = stack.enter_context(
                    TemporaryDirectory(prefix="unearth-download-")
                )
            file = unpack_link(
                self.session,
                link,
                pathlib.Path(download_dir),
                pathlib.Path(location),
                hashes,
                verbosity=self.verbosity,
                download_reporter=download_reporter,
                unpack_reporter=unpack_reporter,
            )
        return file.joinpath(link.subdirectory) if link.subdirectory else file



from unittest import mock

from unearth.utils import LazySequence


def test_lazy_sequence():
    func = mock.Mock()

    def gen(size):
        for i in range(size):
            func()
            yield i

    seq = LazySequence(gen(5))
    assert bool(seq) is True
    assert func.call_count == 1
    assert seq[0] == 0
    assert func.call_count == 1
    assert seq[1] == 1
    assert func.call_count == 2
    assert 3 in seq
    assert func.call_count == 4
    assert len(seq) == 5
    assert list(seq) == [0, 1, 2, 3, 4]
    assert func.call_count == 5



"""Unpack the link to an installed wheel or source."""

from __future__ import annotations

import functools
import hashlib
import logging
import mimetypes
import os
import shutil
import stat
import tarfile
import zipfile
from pathlib import Path
from typing import TYPE_CHECKING, Iterable, cast

import httpx

from unearth.errors import HashMismatchError, UnpackError
from unearth.fetchers import Fetcher
from unearth.link import Link
from unearth.utils import (
    BZ2_EXTENSIONS,
    TAR_EXTENSIONS,
    XZ_EXTENSIONS,
    ZIP_EXTENSIONS,
    display_path,
    format_size,
    iter_with_callback,
)
from unearth.vcs import vcs_support

HTTPErrors: tuple[type[Exception], ...] = (httpx.HTTPError,)
try:
    from requests import HTTPError

    HTTPErrors += (HTTPError,)
except ModuleNotFoundError:
    pass

if TYPE_CHECKING:
    from typing import Protocol

    class DownloadReporter(Protocol):
        def __call__(self, link: Link, completed: int, total: int | None) -> None: ...

    class UnpackReporter(Protocol):
        def __call__(
            self, filename: Path, completed: int, total: int | None
        ) -> None: ...


def noop_download_reporter(link: Link, completed: int, total: int | None) -> None:
    pass


def noop_unpack_reporter(filename: Path, completed: int, total: int | None) -> None:
    pass


READ_CHUNK_SIZE = 8192
logger = logging.getLogger(__name__)


def set_extracted_file_to_default_mode_plus_executable(path: str) -> None:
    """
    Make file present at path have execute for user/group/world
    (chmod +x) is no-op on windows per python docs
    """
    os.chmod(path, (0o777 & ~os.umask(0) | 0o111))


def zip_item_is_executable(info: zipfile.ZipInfo) -> bool:
    mode = info.external_attr >> 16
    # if mode and regular file and any execute permissions for
    # user/group/world?
    return bool(mode and stat.S_ISREG(mode) and mode & 0o111)


def is_within_directory(directory: str | Path, path: str | Path) -> bool:
    try:
        Path(path).relative_to(directory)
    except ValueError:
        return False
    return True


def split_leading_dir(path: str) -> list[str]:
    path = path.lstrip("/").lstrip("\\")
    if "/" in path and (
        ("\\" in path and path.find("/") < path.find("\\")) or "\\" not in path
    ):
        return path.split("/", 1)
    elif "\\" in path:
        return path.split("\\", 1)
    else:
        return [path, ""]


def has_leading_dir(paths: Iterable[str]) -> bool:
    """Returns true if all the paths have the same leading path name
    (i.e., everything is in one subdirectory in an archive)"""
    common_prefix = None
    for path in paths:
        prefix, _ = split_leading_dir(path)
        if not prefix:
            return False
        elif common_prefix is None:
            common_prefix = prefix
        elif prefix != common_prefix:
            return False
    return True


class HashValidator:
    """Validate the hashes of a file."""

    def __init__(self, package_link: Link, hashes: dict[str, list[str]] | None) -> None:
        if hashes is not None:
            # Always sort the hash values for better comparison.
            hashes = {k: sorted(value) for k, value in hashes.items()}
        self.allowed = hashes
        self.package_link = package_link
        self.got = {}
        if hashes is not None:
            for name in hashes:
                try:
                    self.got[name] = hashlib.new(name)
                except (TypeError, ValueError):
                    raise UnpackError(f"Unknown hash name: {name!r}") from None

    def update(self, chunk: bytes) -> None:
        for hasher in self.got.values():
            hasher.update(chunk)

    def validate(self) -> None:
        if not self.allowed:
            return
        gots: dict[str, str] = {}
        for name, hash_list in self.allowed.items():
            got = self.got[name].hexdigest()
            if got in hash_list:
                return
            gots[name] = got
        raise HashMismatchError(self.package_link, self.allowed, gots)

    def validate_path(self, path: Path) -> None:
        with path.open("rb") as f:
            for chunk in iter(lambda: f.read(READ_CHUNK_SIZE), b""):
                self.update(chunk)
        self.validate()


def _check_downloaded(path: Path, hashes: dict[str, list[str]] | None) -> bool:
    """Check if the file has been downloaded."""
    if not path.is_file():
        return False
    try:
        HashValidator(Link.from_path(path), hashes).validate_path(path)
    except HashMismatchError:
        logger.debug("File exists at %s, but the hashes don't match", path)
        path.unlink()
        return False
    logger.debug("The file is already downloaded: %s", path)
    return True


def unpack_archive(
    archive: Path, dest: Path, reporter: UnpackReporter = noop_unpack_reporter
) -> None:
    content_type = mimetypes.guess_type(str(archive))[0]
    if (
        content_type == "application/zip"
        or zipfile.is_zipfile(archive)
        or archive.suffix.lower() in ZIP_EXTENSIONS
    ):
        _unzip_archive(archive, dest, reporter=reporter)
    elif (
        content_type == "application/x-gzip"
        or tarfile.is_tarfile(archive)
        or archive.suffix.lower() in (TAR_EXTENSIONS + XZ_EXTENSIONS + BZ2_EXTENSIONS)
    ):
        _untar_archive(archive, dest, reporter=reporter)
    else:
        raise UnpackError(f"Unknown archive type: {archive.name}")


def _unzip_archive(filename: Path, location: Path, reporter: UnpackReporter) -> None:
    os.makedirs(location, exist_ok=True)
    zipfp = open(filename, "rb")
    with zipfile.ZipFile(zipfp, allowZip64=True) as zip:
        leading = has_leading_dir(zip.namelist())
        callback = functools.partial(reporter, filename, total=len(zip.infolist()))
        for info in iter_with_callback(zip.infolist(), callback):
            name = info.filename
            fn = name
            if leading:
                fn = split_leading_dir(name)[1]
            fn = os.path.join(location, fn)
            dir = os.path.dirname(fn)
            if not is_within_directory(location, fn):
                message = (
                    f"The zip file ({filename}) has a file ({fn}) trying to install "
                    f"outside target directory ({location})"
                )
                raise UnpackError(message)
            if fn.endswith("/") or fn.endswith("\\"):
                # A directory
                os.makedirs(fn, exist_ok=True)
            else:
                os.makedirs(dir, exist_ok=True)
                # Don't use read() to avoid allocating an arbitrarily large
                # chunk of memory for the file's content
                with zip.open(name) as fp, open(fn, "wb") as destfp:
                    shutil.copyfileobj(fp, destfp)

                if zip_item_is_executable(info):
                    set_extracted_file_to_default_mode_plus_executable(fn)


def _untar_archive(filename: Path, location: Path, reporter: UnpackReporter) -> None:
    """Untar the file (with path `filename`) to the destination `location`."""
    os.makedirs(location, exist_ok=True)
    lower_fn = str(filename).lower()
    if lower_fn.endswith(".gz") or lower_fn.endswith(".tgz"):
        mode = "r:gz"
    elif lower_fn.endswith(BZ2_EXTENSIONS):
        mode = "r:bz2"
    elif lower_fn.endswith(XZ_EXTENSIONS):
        mode = "r:xz"
    elif lower_fn.endswith(".tar"):
        mode = "r"
    else:
        logger.warning(
            "Cannot determine compression type for file %s",
            filename,
        )
        mode = "r:*"
    with tarfile.open(filename, mode, encoding="utf-8") as tar:
        leading = has_leading_dir([member.name for member in tar.getmembers()])
        callback = functools.partial(reporter, filename, total=len(tar.getmembers()))
        for member in iter_with_callback(tar.getmembers(), callback):
            fn = member.name
            if leading:
                fn = split_leading_dir(fn)[1]
            path = os.path.join(location, fn)
            if not is_within_directory(location, path):
                message = (
                    f"The tar file ({filename}) has a file ({path}) trying to install "
                    f"outside target directory ({location})"
                )
                raise UnpackError(message)
            if member.isdir():
                os.makedirs(path, exist_ok=True)
            elif member.issym():
                try:
                    tar._extract_member(member, path)
                except Exception as exc:
                    # Some corrupt tar files seem to produce this
                    # (specifically bad symlinks)
                    logger.warning(
                        "In the tar file %s the member %s is invalid: %s",
                        filename,
                        member.name,
                        exc,
                    )
                    continue
            else:
                try:
                    fp = tar.extractfile(member)
                except (KeyError, AttributeError) as exc:
                    # Some corrupt tar files seem to produce this
                    # (specifically bad symlinks)
                    logger.warning(
                        "In the tar file %s the member %s is invalid: %s",
                        filename,
                        member.name,
                        exc,
                    )
                    continue
                os.makedirs(os.path.dirname(path), exist_ok=True)
                assert fp is not None
                with open(path, "wb") as destfp:
                    shutil.copyfileobj(fp, destfp)
                fp.close()
                # Update the timestamp (useful for cython compiled files)
                tar.utime(member, path)
                # member have any execute permissions for user/group/world?
                if member.mode & 0o111:
                    set_extracted_file_to_default_mode_plus_executable(path)


def unpack_link(
    session: Fetcher,
    link: Link,
    download_dir: Path,
    location: Path,
    hashes: dict[str, list[str]] | None = None,
    verbosity: int = 0,
    download_reporter: DownloadReporter = noop_download_reporter,
    unpack_reporter: UnpackReporter = noop_unpack_reporter,
) -> Path:
    """Unpack link into location.

    The link can be a VCS link or a file link.

    Args:
        session (Fetcher): the requests session
        link (Link): the link to unpack
        download_dir (Path): the directory to download the file to
        location (Path): the destination directory
        hashes (dict[str, list[str]]|None): Optional hash dict for validation
        progress_bar (bool): whether to show the progress bar

    Returns:
        Path: the path to the unpacked file or directory
    """
    location.parent.mkdir(parents=True, exist_ok=True)
    if link.is_vcs:
        backend = vcs_support.get_backend(cast(str, link.vcs), verbosity=verbosity)
        download_reporter(link, 0, 1)
        backend.fetch(link, location)
        download_reporter(link, 1, 1)
        return location

    validator = HashValidator(link, hashes)
    if link.is_file:
        if link.file_path.is_dir():
            logger.info(
                "The file %s is a local directory, use it directly",
                display_path(link.file_path),
            )
            return link.file_path
        artifact = link.file_path
        validator.validate_path(artifact)
    else:
        # A remote artfiact link, check the download dir first
        artifact = download_dir / link.filename
        if not _check_downloaded(artifact, hashes):
            with session.get_stream(link.normalized) as resp:
                try:
                    resp.raise_for_status()
                except HTTPErrors as e:
                    raise UnpackError(f"Download failed: {e}") from None
                try:
                    total = int(resp.headers["Content-Length"])
                except (KeyError, ValueError, TypeError):
                    total = None
                if getattr(resp, "from_cache", False):
                    logger.info("Using cached %s", link)
                else:
                    size = format_size(resp.headers.get("Content-Length", ""))
                    logger.info("Downloading %s (%s)", link, size)
                with artifact.open("wb") as f:
                    callback = functools.partial(download_reporter, link, total=total)
                    for chunk in iter_with_callback(
                        resp.iter_bytes(chunk_size=READ_CHUNK_SIZE),
                        callback,
                        stepper=len,
                    ):
                        if chunk:
                            validator.update(chunk)
                            f.write(chunk)
            validator.validate()
    if link.is_wheel:
        if link.is_file:
            # Use the local file directly
            return artifact
        target_file = location / link.filename
        if target_file != artifact:
            # For wheels downloaded from remote locations, move it to the destination.
            os.replace(artifact, target_file)
        return target_file

    unpack_archive(artifact, location, reporter=unpack_reporter)
    return location



from __future__ import annotations

import abc
import logging
import os
import shutil
import subprocess
from pathlib import Path
from typing import Collection, Sequence, Type, TypeVar, cast

from unearth.errors import UnpackError, URLError, VCSBackendError
from unearth.link import Link
from unearth.utils import compare_urls

logger = logging.getLogger(__name__)


class HiddenText:
    """A string that redacts the auth info from the URL."""

    def __init__(self, secret: str, redacted: str) -> None:
        self.secret = secret
        self.redacted = redacted

    def __str__(self) -> str:
        return self.redacted

    def __repr__(self) -> str:
        return f"<URL {str(self)!r}>"


class VersionControl(abc.ABC):
    """The base class for all version control systems.

    Attributes:
        name: the backend name
        dir_name: the backend data directory, such as '.git'
        action: the word to describe the clone action.
    """

    name: str
    dir_name: str

    def __init__(self, verbosity: int = 0) -> None:
        self.verbosity = verbosity

    def run_command(
        self,
        cmd: Sequence[str | HiddenText],
        cwd: Path | None = None,
        extra_env: dict[str, str] | None = None,
        log_output: bool = True,
        stdout_only: bool = False,
        extra_ok_returncodes: Collection[int] = (),
    ) -> subprocess.CompletedProcess[str]:
        """Run the command in the given working directory."""
        env = None
        if extra_env:
            env = dict(os.environ, **extra_env)
        try:
            cmd = [self.name, *cmd]
            display_cmd = subprocess.list2cmdline(map(str, cmd))
            logger.debug("Running command %s", display_cmd)
            result = subprocess.run(
                [v.secret if isinstance(v, HiddenText) else v for v in cmd],
                cwd=str(cwd) if cwd else None,
                check=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.DEVNULL if stdout_only else subprocess.STDOUT,
                env=env,
                text=True,
            )
        except subprocess.CalledProcessError as e:
            if e.returncode in extra_ok_returncodes:
                if log_output:
                    logger.debug(e.stdout.rstrip())
                return subprocess.CompletedProcess(e.args, e.returncode, e.stdout)
            raise UnpackError(e.output) from None
        except FileNotFoundError:
            logger.debug(f"Cannot find `{self.name}`, PATH={os.environ.get('PATH')}")
            msg = (
                f"Unable to find executable `{self.name}`, "
                "make sure it's installed in PATH."
            )
            raise FileNotFoundError(msg) from None
        else:
            if log_output:
                logger.debug(result.stdout.rstrip())
            return result

    def _is_local_repository(self, repo: str) -> bool:
        """
        posix absolute paths start with os.path.sep,
        win32 ones start with drive (like c:\\folder)
        """
        drive, _ = os.path.splitdrive(repo)
        return repo.startswith(os.path.sep) or bool(drive)

    def get_url_and_rev_options(
        self, link: Link
    ) -> tuple[HiddenText, str | None, list[str | HiddenText]]:
        """Get the URL and revision options from the link."""
        parsed = link.parsed
        scheme = parsed.scheme.rsplit("+", 1)[-1]
        netloc, user, password = self.get_netloc_and_auth(parsed.netloc, scheme)
        if password is not None:
            password = HiddenText(password, "***")  # type: ignore[assignment]
        replace_dict = {
            "scheme": parsed.scheme.rsplit("+", 1)[-1],
            "netloc": netloc,
            "fragment": "",
        }
        if "@" not in parsed.path:
            rev = None
        else:
            path, _, rev = parsed.path.rpartition("@")
            if not rev:
                raise URLError(
                    f"The url {link.redacted!r} has an empty revision (after @)."
                    "You should specify a revision or remove the @ from the URL."
                )
            replace_dict["path"] = path
        args = self.make_auth_args(user, cast(HiddenText, password))
        url = parsed._replace(**replace_dict).geturl()
        hidden_url = HiddenText(url, Link(url).redacted)
        return hidden_url, rev, args

    def fetch(self, link: Link, location: Path) -> None:
        """Clone the repository to the destination directory, and return
        the path to the local repository.

        Args:
            link (Link): the VCS link to the repository
            location (Path): the destination directory
        """
        url, rev, args = self.get_url_and_rev_options(link)
        if not location.exists():
            return self.fetch_new(location, url, rev, args)

        if not self.is_repository_dir(location) or not compare_urls(
            url.secret, self.get_remote_url(location)
        ):
            if not self.is_repository_dir(location):
                logger.debug(f"{location} is not a repository directory, removing it.")
            else:
                remote_url = self.get_remote_url(location)
                logger.debug(
                    f"{location} is a repository directory, but the remote url "
                    f"{remote_url!r} does not match the url {url!r}."
                )
            shutil.rmtree(location)
            return self.fetch_new(location, url, rev, args)

        if self.is_commit_hash_equal(location, rev):
            logger.debug("Repository %s is already up-to-date", location)
            return
        self.update(location, rev, args)

    @abc.abstractmethod
    def fetch_new(
        self,
        location: Path,
        url: HiddenText,
        rev: str | None,
        args: list[str | HiddenText],
    ) -> None:
        """Fetch the repository from the remote link, as if it is the first time.

        Args:
            location (Path): the repository location
            link (Link): the VCS link to the repository
            rev (str|None): the revision to checkout
            args (list[str | HiddenText]): the arguments to pass to the update command
        """
        pass

    @abc.abstractmethod
    def update(
        self, location: Path, rev: str | None, args: list[str | HiddenText]
    ) -> None:
        """Update the repository to the given revision.

        Args:
            location (Path): the repository location
            rev (str|None): the revision to checkout
            args (list[str | HiddenText]): the arguments to pass to the update command
        """
        pass

    @abc.abstractmethod
    def get_remote_url(self, location: Path) -> str:
        """Get the remote URL of the repository."""
        return ""

    @abc.abstractmethod
    def get_revision(self, location: Path) -> str:
        """Get the commit hash of the repository."""
        pass

    def is_immutable_revision(self, location: Path, link: Link) -> bool:
        """Check if the revision is immutable.
        Always return False if the backend doesn't support immutable revisions.
        """
        return False

    def get_rev_args(self, rev: str | None) -> list[str]:
        """Get the revision arguments for the command."""
        return [rev] if rev is not None else []

    def is_commit_hash_equal(self, location: Path, rev: str | None) -> bool:
        """Always assume the versions don't match"""
        return False

    def is_repository_dir(self, location: Path) -> bool:
        """Check if the given directory is a repository directory."""
        return location.joinpath(self.dir_name).exists()

    def get_netloc_and_auth(
        self, netloc: str, scheme: str
    ) -> tuple[str, str | None, str | None]:
        """Get the auth info and the URL from the link.
        For VCS like git, the auth info must stay in the URL.
        """
        return netloc, None, None

    def make_auth_args(
        self, user: str | None, password: HiddenText | None
    ) -> list[str | HiddenText]:
        """Make the auth args for the URL."""
        return []


_V = TypeVar("_V", bound=Type[VersionControl])


class VcsSupport:
    def __init__(self) -> None:
        self._registry: dict[str, Type[VersionControl]] = {}

    def register(self, vcs: _V) -> _V:
        self._registry[vcs.name] = vcs
        return vcs

    def unregister_all(self) -> None:
        self._registry.clear()

    def get_backend(self, name: str, verbosity: int = 0) -> VersionControl:
        try:
            return self._registry[name](verbosity=verbosity)
        except KeyError:
            raise VCSBackendError(name) from None


vcs_support = VcsSupport()



"""The command-line interface for the unearth package."""

from __future__ import annotations

import argparse
import json
import logging
import os
import sys
import tempfile
from dataclasses import dataclass

from packaging.requirements import Requirement

from unearth.evaluator import TargetPython
from unearth.finder import PackageFinder
from unearth.link import Link
from unearth.utils import splitext


@dataclass(frozen=True)
class CLIArgs:
    requirement: Requirement
    verbose: bool
    index_urls: list[str]
    find_links: list[str]
    trusted_hosts: list[str]
    no_binary: bool
    only_binary: bool
    prefer_binary: bool
    all: bool
    link_only: bool
    download: str | None
    py_ver: tuple[int, ...] | None
    abis: list[str] | None
    impl: str | None
    platforms: list[str] | None


def _setup_logger(verbosity: bool) -> None:
    logger = logging.getLogger("unearth")
    logger.setLevel(logging.DEBUG if verbosity else logging.WARNING)
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter("%(levelname)s: %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)


def comma_split(arg: str) -> list[str]:
    return arg.split(",")


def to_py_ver(arg: str) -> tuple[int, ...]:
    return tuple(int(i) for i in arg.split(".") if i.isdigit())


def cli_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Find and download packages from a PEP 508 requirement string.",
    )
    parser.add_argument(
        "requirement",
        type=Requirement,
        help="A PEP 508 requirement string, e.g. 'requests>=2.18.4'.",
    )
    parser.add_argument(
        "--verbose", "-v", action="store_true", help="Enable verbose logging."
    )
    parser.add_argument(
        "--index-url",
        "-i",
        metavar="URL",
        dest="index_urls",
        action="append",
        help="(Multiple)(PEP 503)Simple Index URLs.",
    )
    parser.add_argument(
        "--find-link",
        "-f",
        dest="find_links",
        metavar="LOCATION",
        action="append",
        help="(Multiple)URLs or locations to find links from.",
    )
    parser.add_argument(
        "--trusted-host",
        dest="trusted_hosts",
        metavar="HOST",
        action="append",
        help="(Multiple)Trusted hosts that should skip the verification.",
    )
    parser.add_argument(
        "--no-binary",
        action="store_true",
        help="Exclude binary packages from the results.",
    )
    parser.add_argument(
        "--only-binary",
        action="store_true",
        help="Only include binary packages in the results.",
    )
    parser.add_argument(
        "--prefer-binary",
        action="store_true",
        help="Prefer binary packages even if sdist candidates of newer versions exist.",
    )
    parser.add_argument(
        "--all", action="store_true", help="Return all applicable versions."
    )
    parser.add_argument(
        "--link-only",
        "-L",
        action="store_true",
        help="Only return links instead of a JSON object.",
    )
    parser.add_argument(
        "--download",
        "-d",
        nargs="?",
        const=".",
        metavar="DIR",
        help="Download the package(s) to DIR.",
    )
    group = parser.add_argument_group("Target Python options")
    group.add_argument(
        "--python-version",
        "--py",
        dest="py_ver",
        type=to_py_ver,
        help="Target Python version. e.g. 3.11.0",
    )
    group.add_argument(
        "--abis", type=comma_split, help="Comma-separated list of ABIs. e.g. cp39,cp310"
    )
    group.add_argument(
        "--implementation",
        "--impl",
        dest="impl",
        help="Python implementation. e.g. cp,pp,jy,ip",
    )
    group.add_argument(
        "--platforms",
        type=comma_split,
        help="Comma-separated list of platforms. e.g. win_amd64,linux_x86_64",
    )
    return parser


def get_dest_for_package(dest: str, link: Link) -> str:
    if link.is_wheel:
        return dest
    filename = link.filename.rsplit("@", 1)[0]
    fn, _ = splitext(filename)
    return os.path.join(dest, fn)


def cli(argv: list[str] | None = None) -> None:
    parser = cli_parser()
    args = CLIArgs(**vars(parser.parse_args(argv)))
    _setup_logger(args.verbose)
    name = args.requirement.name
    target_python = TargetPython(args.py_ver, args.abis, args.impl, args.platforms)
    finder = PackageFinder(
        index_urls=args.index_urls or [],
        find_links=args.find_links or [],
        trusted_hosts=args.trusted_hosts or [],
        target_python=target_python,
        no_binary=[name] if args.no_binary else [],
        only_binary=[name] if args.only_binary else [],
        prefer_binary=[name] if args.prefer_binary else [],
        verbosity=int(args.verbose),
    )
    matches = list(finder.find_matches(args.requirement))
    if not matches:
        print("No matches are found.", file=sys.stderr)
        sys.exit(1)
    if not args.all:
        matches = matches[:1]

    result = []
    if args.download:
        os.makedirs(args.download, exist_ok=True)
    with tempfile.TemporaryDirectory("unearth-download-") as download_dir:
        for match in matches:
            data = match.as_json()
            if args.download is not None:
                dest = get_dest_for_package(args.download, match.link)
                data["local_path"] = finder.download_and_unpack(
                    match.link,
                    dest,
                    download_dir,
                ).as_posix()
            result.append(data)
    if args.link_only:
        for item in result:
            print(item["link"]["url"])
            if "local_path" in item:
                print("  ==>", item["local_path"])
    else:
        print(json.dumps(result[0] if len(result) == 1 else result, indent=2))


if __name__ == "__main__":
    cli()
